import numpy as np
from .calc import CalcV, CalcD
from tqdm import tqdm
import dill


class Solver:

    def __init__(self, l=None, x=None, t=None, rho0=None, c=None, boundary=None, diff=False, potential=None):
        '''
        constructor of the class
        Parameters
        ----------
        l : list or numpy array
        x : list or numpy array
        t : list or numpy array
        rho0 : function or tuple of functions
        c : float or tuple of floats
        boundary : None or tuple
        diff : bool
        '''
        self.l, self.x, self.t = self.correct_l_x_t(l, x, t)
        self.c, self.rho0 = self.to_tuple(c, rho0)
        self.boundary = self.correct_boundary(boundary)
        self.potential = self.get_potential(potential)
        self.int_x, self.int_t, self.int_l, self.steps_number = self.get_ints()
        self.diff = diff
        self.convergence = []
        self.grid = self.create_initial_grid()

    def correct_l_x_t(self, l, x, t):
        '''
        checking correctnes of the input and changes list into the numpy arrays
        Parameters
        ----------
        l : list or numpy array
        x : list or numpy array
        t : list or numpy array

        Returns
        -------
        l, x, t as numpy arrays
        '''
        if isinstance(x, np.ndarray) == False:
            x = np.array(x)

        if isinstance(t, np.ndarray) == False:
            t = np.array(t)

        if isinstance(l, np.ndarray) == False:
            l = np.array(l)

        if t.size < 2:
            raise ValueError('t grid size cannot be smaller than 2')

        if l.ndim != 1 or x.ndim != 1 or t.ndim != 1:
            raise ValueError('x, l and t should be 1D arrays')

        return l, x, t

    def correct_boundary(self, boundary):
        '''
        checks if boundary is inputted correctly
        Parameters
        ----------
        boundary : None or tuple

        Returns
        -------
        Raises error or unchanged boundary
        '''
        if boundary is None:
            return None

        if isinstance(boundary, tuple) == False or len(boundary) != 2:
            raise TypeError('boundary argument should be a tuple of lenght 2')

        return boundary

    def get_potential(self, potential):

        if potential is not None:
            potential = potential(self.x)[np.newaxis, np.newaxis, :]

        return potential

    def to_tuple(self, c, rho0):
        '''
        ensures that c and rho0 are tuple and have this same size
        Parameters
        ----------
        c
        rho0

        Returns
        -------
        c and rho0 as tuple
        '''
        if isinstance(c, tuple) == False:
            c = (c,)

        if isinstance(rho0, tuple) == False:
            rho0 = (rho0,)

        if len(c) != len(rho0):
            raise TypeError('c and rho0 have to have this same size')

        return c, rho0

    def get_ints(self):
        '''
        extracts required data from x,t grids
        Returns
        -------
        int_x, int_l, int_t : floats representing spacing in dimensions
        steps_number : int, number of time iterations
        '''
        int_x = np.diff(self.x).mean()

        int_t = np.diff(self.t).mean()

        int_l = np.diff(self.l).mean()

        steps_number = self.t.size

        return int_x, int_t, int_l, steps_number

    # creates grid filled with initial condition
    def create_initial_grid(self):
        '''
        Returns
        -------
        numpy array which is solution grid filled only with initial state
        '''
        # dimensions: N, l, x, t

        grid = np.zeros((len(self.c), self.l.size, self.x.size, self.t.size))

        N_dim = []

        for item in self.rho0:
            N = item(self.l[:, np.newaxis], self.x[np.newaxis, :])

            N_dim.append(N)

        initial_state = np.stack(N_dim)

        grid[Ellipsis, 0] = initial_state

        return grid

    def create_matrix(self, time):
        '''

        Parameters
        ----------
        time : int

        Returns
        -------
        numpy array which is a matrix required for implicit solving of GHD
        '''
        # dimensions N, l, x for rho and V

        rho = self.grid[Ellipsis, time]

        # calculating V using class CalcV

        V = CalcV(rho, self.l, self.c).V

        # changing indices back to proper order

        V = np.einsum('Nxl -> Nlx', V)

        h = self.int_t / (2 * self.int_x)

        V_V_h = h * V[Ellipsis, np.newaxis, :] * np.ones_like(V)[Ellipsis, np.newaxis]

        off_diagonal = np.zeros((V_V_h.shape[-1], V_V_h.shape[-1]))

        np.fill_diagonal(off_diagonal[:, 1:], 1)
        np.fill_diagonal(off_diagonal[1:, 0:], -1)

        off_diagonal = V_V_h * off_diagonal[np.newaxis, np.newaxis, Ellipsis]

        diagonal = np.zeros((V_V_h.shape[-1], V_V_h.shape[-1]))

        np.fill_diagonal(diagonal, 1)

        diagonal = np.ones_like(V_V_h) * diagonal[np.newaxis, np.newaxis, Ellipsis]

        matrix = diagonal + off_diagonal

        if self.boundary is None:

            matrix[Ellipsis, 0, -1] = -V_V_h[Ellipsis, 0, -1]

            matrix[Ellipsis, -1, 0] = V_V_h[Ellipsis, -1, 0]

        else:

            matrix[Ellipsis, 0, 0] = self.boundary[0]

            matrix[Ellipsis, 0, 1] = 0

            matrix[Ellipsis, -1, -1] = self.boundary[1]
            matrix[Ellipsis, -1, -2] = 0

        return matrix

    def x_der(self, arr):
        '''

        Parameters
        ----------
        arr : numpy array of the grid at given time

        Returns
        -------
        numpy array which is a derivative in x dimension of inputted arr
        '''
        der = 1 / (2 * self.int_x) * (np.roll(arr, -1, axis=-1) - np.roll(arr, 1, axis=-1))

        # der = np.gradient(arr, self.int_x, axis = -1)

        return der

    def lambda_der(self, arr):
        '''

        Parameters
        ----------
        arr : numpy array of the grid at given time

        Returns
        -------
        numpy array which is a derivative in l dimension of inputted arr
        '''
        der = 1 / (2 * self.int_l) * (np.roll(arr, -1, axis=-2) - np.roll(arr, 1, axis=-2))

        return der

    def diff_fixed_point_func(self, rho, rho_next, D, V):
        '''
        functions utilizing fixed point iteration method for solving GHD with diffusion
        Parameters
        ----------
        rho : numpy array, state for which it is calculated
        rho_next
        D : numpy  array diffusion operator
        V : numpy array, effective velocity

        Returns
        -------

        '''
        diff = rho_next

        # D, V dimensions N, x, momenta

        D_op = np.einsum('Nxos, Nsx -> Nox', D, self.x_der(rho_next), optimize=True)

        V_rho = np.einsum('Nxl, Nlx -> Nlx', V, rho_next, optimize=True)

        if self.potential is None:

            foo = rho + self.int_t * (self.x_der(D_op) / 2 - self.x_der(V_rho))

        else:

            foo = rho + self.int_t * (
                    self.x_der(D_op) / 2 - self.x_der(V_rho) + self.x_der(self.potential) * self.lambda_der(rho))

        diff = np.abs(diff - foo).mean()

        return foo, diff

    def diff_equ(self, time):
        '''
        solves GHD at given time with diffusion using fixed point iteration method
        Parameters
        ----------
        time : int

        Returns
        -------
        rho_next: numpy array, state at another time
        diff : list, collects convergence of the method
        '''
        rho_next = self.grid[Ellipsis, time]

        rho = self.grid[Ellipsis, time]

        Diff = CalcD(rho, self.l, self.c)

        D, V = Diff.D, Diff.V  # dimensions N, x, momenta

        diff = []

        for i in range(15):
            rho_next, difference = self.diff_fixed_point_func(rho, rho_next, D, V)

            diff.append(difference)

        return rho_next, diff

    def solve_equation(self, path=None):
        '''
        Function solving GHD or GHD with diffusion using previous methods
        Parameters
        ----------
        path : optional string, if is not None, Solver objects is saved to localization

        Returns

        numpy array which is solution to GHD equations
        -------

        '''
        if self.diff is False:

            for time in tqdm(range(self.t.size - 1)):
                matrix = self.create_matrix(time)

                self.grid[Ellipsis, time + 1] = np.linalg.solve(matrix, self.grid[Ellipsis, time])

        else:

            for time in tqdm(range(self.t.size - 1)):
                rho_next, diff = self.diff_equ(time)

                self.grid[Ellipsis, time + 1] = rho_next

                self.convergence.append(diff)

        if path is not None:
            with open(path, 'wb') as file:
                dill.dump(self, file)

    def save_array(self, path):
        '''
        Saves array of the solutin
        Parameters
        ----------
        path : string, location to which array is saved

        '''
        with open(path, 'wb') as file:
            dill.dump(self.grid, file)

    def save(self, path):
        '''
        Saves Solver object as binary dill file to path localization
        Parameters
        ----------
        path : string, location to which object is saved
        '''
        with open(path, 'wb') as file:
            dill.dump(self, file)

            # loading arrays

    @staticmethod
    def load(path):
        '''
        Loads array or Solver object from the binary file
        Parameters
        ----------
        path : string, localizaction of binary file

        Returns
        -------
        Solver object from the path localization
        '''
        with open(path, 'rb') as file:
            arr = dill.load(file)

        return arr



import numpy as np


class CalcV:
    '''
    Class calculating effective velocity of given state rho
    '''

    def __init__(self, rho, l, c):
        '''
        Constructor of a class
        Parameters
        ----------
        rho : numpy array, state
        l : numpy array, momenta
        c : tuple of floats, coupling constants
        '''
        # dimensions of rho N, x, l
        self.rho = self.get_rho(rho)
        self.l = l
        self.c = c
        self.int_l = np.diff(self.l).mean()
        self.T = self.create_T()
        self.n, self.rho_tot = self.get_n_rho_tot()
        self.operator = self.get_operator()
        self.V = self.get_V()

    def get_rho(self, rho):
        '''
        Changes dimensions order, to more useful in further calculations
        Parameters
        ----------
        rho : numpy array

        Returns
        -------
        numpy array, rho with dimensions: N, x, l

        '''
        rho = np.einsum('Nlx -> Nxl', rho)

        return rho

    def create_T(self):
        '''
        Returns
        -------
        numpy array, integral kernel of GHD
        '''
        l, u = np.meshgrid(self.l, self.l, indexing='ij')

        T_grid = []

        for value in self.c:
            T = value / np.pi * 1 / ((l - u) ** 2 + value ** 2)

            T_grid.append(T)

        # dimensions N, l, u   
        T = np.stack(T_grid)

        return T

    def get_n_rho_tot(self):
        '''
        Calculates n and rho total
        Returns
        -------
        n : numpy array
        rho_tot : numpy array
        '''
        # new indices are rho: N, x, l

        rho_tot = 1 / (2 * np.pi) + np.einsum('Nlu, Nxu -> Nxl', self.T, self.rho) * self.int_l

        n = self.rho / rho_tot

        return n, rho_tot

    def get_operator(self):
        '''
        Creates 1 -Tn operator required for velocity calculations
        Returns
        -------
        operator : numpy array
        '''

        # dimensions : T(N,l,u) , n(N, x, u)
        Tn = self.T[:, np.newaxis, :, :] * self.n[:, :, np.newaxis, :]

        # create delta l,u for each N and x
        # dimensions N, x, l, u

        delta = np.identity(self.l.size)[np.newaxis, np.newaxis, :, :]

        operator = delta - Tn * self.int_l

        # dimensions N,x, l, u
        operator = np.linalg.inv(operator)

        return operator

    def get_V(self):
        '''
        Calculates effective velocity
        Returns
        -------
        V : numpy array
        '''
        # dimensions N, x, l

        u = 2 * self.l

        k_dr = np.sum(self.operator, axis=-1)

        omega_dr = np.einsum('Nxlu, u -> Nxl', self.operator, u)

        V = omega_dr / k_dr

        return V


class CalcD(CalcV):
    '''
    Class calculating diffusion operator for given state rho, derived class of CalcV
    '''

    def __init__(self, rho, l, c):
        '''
        Constructor of this class
        Parameters
        ----------
        rho : numpy array, state at given time
        l : numpy array, momenta dimension
        c : tuple of floats, coupling constants
        '''
        super().__init__(rho, l, c)

        # dimensions N, x, l    
        self.W = self.get_W()
        self.w = np.sum(self.W, axis=-2) * self.int_l
        # self.w = self.get_w()
        self.D_ker = self.get_D_ker()
        self.D = self.get_D()

    def get_W(self):
        '''
        Calculates W operatos
        Returns
        -------
        W : numpy array
        '''
        T_dr = np.einsum('Nxlu, Nuo -> Nxlo', self.operator, self.T, optimize=True)

        # Now order of indices is N, x, l, u

        rho = self.rho[Ellipsis, np.newaxis]

        n = self.n[Ellipsis, np.newaxis]

        # rho_tot = self.rho_tot[Ellipsis, np.newaxis]

        W = rho * (1 - n) * T_dr ** 2 * np.abs(self.V[Ellipsis, np.newaxis] - self.V[Ellipsis, np.newaxis, :])

        return W

    def get_D_ker(self):
        '''
        Calculates D_ker operator, where D operator is  (1 -Tn)-1 rho D_ker rho-1 (1 -Tn)
        Returns
        -------
        D_ker : numpy array
        '''
        delta = np.identity(self.l.size)[np.newaxis, np.newaxis, Ellipsis]

        rho_tot = self.rho_tot[Ellipsis, np.newaxis]

        # dimensions N, x, l, u
        D_ker = (delta * self.w[Ellipsis, np.newaxis] - self.W * self.int_l) / rho_tot ** 2

        return D_ker

    def get_D(self):
        '''
        Calculates diffusion operator
        Returns
        -------
        D : numpy array
        '''
        Tn = self.T[:, np.newaxis, :, :] * self.n[:, :, np.newaxis, :]

        delta = np.identity(self.l.size)[np.newaxis, np.newaxis, Ellipsis]

        op_ker = delta - Tn * self.int_l

        rho_factor = self.rho_tot[Ellipsis, np.newaxis] / self.rho_tot[Ellipsis, np.newaxis, :]

        D_ker = self.D_ker * rho_factor

        D = np.einsum('Nxou, Nxul , Nxls -> Nxos', self.operator, D_ker, op_ker, optimize=True)

        return D

# Changes:  rho_factor is in different place and D_ker is devided by rho_tot**2



import matplotlib.pyplot as plt
import numpy as np
import dill
import scipy as sci


class Observable:

    def __init__(self, Solver_object):

        self.object = self.get_object(Solver_object)
        # dimensions: N, l, x, t       
        self.T = self.get_T()
        self.rho_tot = self.get_rho_tot()
        self.rho_h = self.get_rho_h()
        # self.n = self.get_n()

    def get_object(self, inp) -> object:
        '''
        loading Solver object into observable class
        Parameters
        ----------
        inp

        Returns
        -------

        '''
        if isinstance(inp, str):

            with open(inp, 'rb') as file:
                arr = dill.load(file)

            return arr

        else:
            return inp

    def get_T(self):

        # numba doesn't support meshgrid
        l, u = np.meshgrid(self.object.l, self.object.l, indexing='ij')

        T_grid = []

        for value in self.object.c:
            T = value / np.pi * 1 / ((l - u) ** 2 + value ** 2)

            T_grid.append(T)

        # dimensions N, l, u   
        T = np.stack(T_grid)

        return T

    def get_rho_tot(self):

        # new indices are rho: N, x, l

        rho_tot = 1 / (2 * np.pi) + np.einsum('Nlu, Nuxt -> Nlxt', self.T, self.object.grid,
                                              optimize=True) * self.object.int_l

        return rho_tot

    def get_rho_h(self):

        rho_h = self.rho_tot - self.object.grid

        return rho_h

    def energy(self, option='total'):
        # Dimensions N, l, x, t
        l = self.object.l[np.newaxis, :, np.newaxis, np.newaxis]
        energy_grid = self.object.grid * l ** 2

        option_mapping = {
            'local': (np.sum(energy_grid, axis=1) * self.object.int_l),
            'theta': (np.sum(energy_grid, axis=2) * self.object.int_x),
            'total': (np.sum(energy_grid, axis=(1, 2)) * self.object.int_l * self.object.int_x)
        }

        if option not in option_mapping:
            raise ValueError('Incorrect argument, choose between: local, theta, total')

        energy = option_mapping[option]

        if energy.shape[0] == 1:
            energy = energy[0, Ellipsis]

        return energy

    # # here I can make it better if the index is negative
    def n(self, option='total'):
        # Dimensions N, l, x, t

        n_grid = self.object.grid / self.rho_tot

        option_mapping = {
            'local': (np.sum(n_grid, axis=1) * self.object.int_l),
            'theta': (np.sum(n_grid, axis=2) * self.object.int_x),
            'total': (np.sum(n_grid, axis=(1, 2)) * self.object.int_l * self.object.int_x)
        }

        if option not in option_mapping:
            raise ValueError('Incorrect argument, choose between: local, theta, total')

        n = option_mapping[option]

        if n.shape[0] == 1:
            n = n[0, Ellipsis]

        return n

    def entropy(self, option='total'):

        # Dimensions N, l, x, t

        rho = self.object.grid

        rho[rho < 0] = 0

        S_grid = self.rho_tot * np.log(self.rho_tot) - sci.special.xlogy(rho, rho) - sci.special.xlogy(self.rho_h,
                                                                                                       self.rho_h)
        option_mapping = {
            'local': (np.sum(S_grid, axis=1) * self.object.int_l),
            'theta': (np.sum(S_grid, axis=2) * self.object.int_x),
            'total': (np.sum(S_grid, axis=(1, 2)) * self.object.int_l * self.object.int_x)
        }

        if option not in option_mapping:
            raise ValueError('Incorrect argument, choose between: local, theta, total')

        entropy = option_mapping[option]

        if entropy.shape[0] == 1:
            entropy = entropy[0, Ellipsis]

        return entropy

    def plot_template(self, observable, option='local', frames=(0, -1), N=0, path=None, name='', style='-'):

        main_options_dictionairy = {
            'n': {
                'local': {'x_axis': self.object.x, 'y_axis': self.n('local'), 'x_label': 'x', 'y_label': 'n'},
                'theta': {'x_axis': self.object.l, 'y_axis': self.n('theta'), 'x_label': 'theta', 'y_label': 'n'},
                'total': {'x_axis': self.object.t, 'y_axis': self.n('total'), 'x_label': 'time', 'y_label': 'n'}},

            'energy': {
                'local': {'x_axis': self.object.x, 'y_axis': self.energy('local'), 'x_label': 'x', 'y_label': 'energy'},
                'theta': {'x_axis': self.object.l, 'y_axis': self.energy('theta'), 'x_label': 'theta',
                          'y_label': 'energy'},
                'total': {'x_axis': self.object.t, 'y_axis': self.energy('total'), 'x_label': 'time',
                          'y_label': 'energy'}},

            'entropy': {
                'local': {'x_axis': self.object.x, 'y_axis': self.entropy('local'), 'x_label': 'x',
                          'y_label': 'entropy'},
                'theta': {'x_axis': self.object.l, 'y_axis': self.entropy('theta'), 'x_label': 'theta',
                          'y_label': 'entropy'},
                'total': {'x_axis': self.object.t, 'y_axis': self.entropy('total'), 'x_label': 'time',
                          'y_label': 'entropy'}}
        }

        option_mapping = main_options_dictionairy[observable]

        if option not in option_mapping:
            raise ValueError('Incorrect argument, choose between: local, theta or total')

        if self.object.grid.shape[0] == 1:
            y_axis = option_mapping[option]['y_axis']

        else:

            y_axis = option_mapping[option]['y_axis'][N, Ellipsis]

        if option == 'total':
            plt.plot(option_mapping[option]['x_axis'], y_axis, style,
                     )
            plt.xlabel('time')
            plt.ylabel(option_mapping[option]['y_label'])
            plt.show()

        else:
            for item in frames:
                plt.plot(option_mapping[option]['x_axis'], y_axis[:, item], style,
                         label='{} t = {}'.format(name, round(item * self.object.int_t, 3)))
            plt.xlabel(option_mapping[option]['x_label'])
            plt.ylabel(option_mapping[option]['y_label'])
            plt.legend()
            plt.show()

        if path is not None:
            plt.savefig(path)

    def plot_n(self, option='local', frames=(0, -1), N=0, path=None, name='', style='-'):

        self.plot_template(observable='n', option=option, frames=frames, N=N, path=path, name=name, style=style)

    def plot_energy(self, option='total', frames=(0, -1), N=0, path=None, name='', style='-'):

        self.plot_template(observable='energy', option=option, frames=frames, N=N, path=path, name=name, style=style)

    def plot_entropy(self, option='total', frames=(0, -1), N=0, path=None, name='', style='-'):

        self.plot_template(observable='entropy', option=option, frames=frames, N=N, path=path, name=name, style=style)



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="BetheFluid",
    version="0.3.5",
    description='Python package for solving GHD equations',
    author='Antoni Lis, Milosz Panfil',
    packages=['BetheFluid'],
    install_requires=requirements,
    zip_safe=False
)